library(Rcpp)
library(mvtnorm)
library(Matrix)


### set the following user parameters:
## 1) output_filepath: directory containing the code as well as the patientinfo and rds file
## 2) name of the patientinfo file
## 3) name of the rds file containing daily digital phenotypes
output_filepath="C:/Users/Ian/Dropbox/SZpilotAnomalyDetection/code"
setwd(output_filepath)
pinfo = read.csv("patientinfo.csv",stringsAsFactors=F)
infilename=paste("feature_matrix_clean_2daycarry_combined.rds",sep="")
infilepath=paste(output_filepath,infilename,sep="/")




source('TSAnomalyDetection.R')

### 1) Perform anomaly detection

anomaly_detection_plot = function(ID,Nsurveys,NoSurvey=FALSE,vertmarks=NULL,onesided=TRUE){
  infilename=paste("feature_matrix_clean_",daysback,"daycarry_combined.rds",sep="")
  infilepath=paste(output_filepath,infilename,sep="/")
  if(!file.exists(infilepath)){return(NULL)}
  Nmobfeats=15
  if(length(grep("/",vertmarks))>0){
    vertmarks = MDYtoYMD(vertmarks)
  }
  dat = readRDS(infilepath)[[1]]
  outmat = matrix(NA,nrow=nrow(dat),ncol=9)
  #dat = read.table(paste(homedir,"Processed_data",filename,sep="/"),header=T,stringsAsFactors=F)
  # if(ID=="vh7ab23d"){
  #   rIDs=which(dat$IID==ID)[1:(42+0)]
  # }else if(ID=="v77ngnht"){
  #   rIDs=which(dat$IID==ID)[1:(50+0)]
  # }else if(ID=="ety2hu13"){
  #   rIDs=which(dat$IID==ID)[1:(38+0)]
  # }else{
  #   rIDs=which(dat$IID==ID)
  # }
  rIDs=which(dat$IID==ID)
  y = dat[rIDs,]
  if(nrow(y)<4){
    cat(" Not enough data for anomaly detection.")
    return(NULL)
  }
  y = y[c(-1,-nrow(y)),]
  fvar=apply(y[,3:ncol(y)],2,function(xx) var(xx,na.rm=T))
  fext=apply(y[,3:ncol(y)],2,function(xx) length(which(!is.na(xx))))
  IDrm=unique(c(which(is.na(fvar)),which(fvar==0),which(fext<3)))+2
  Nsurvrm=length(which(IDrm<=Nsurveys+2))
  Nmobrm=length(which(IDrm<=Nmobfeats+Nsurveys+2))-Nsurvrm
  Nsocfeats=ncol(y)-Nmobfeats-Nsurveys-2-(length(IDrm)-Nsurvrm-Nmobrm)
  Nsurveys=Nsurveys-Nsurvrm
  Nmobfeats=Nmobfeats-Nmobrm
  if(length(IDrm)>0){
    y = y[,-IDrm]
  }
  if(Nsocfeats+Nsurveys+Nmobfeats==0){
    cat(" Not enough data for anomaly detection.")
    return(NULL)
  }
  for(i in 1:3){
    if(i==1){
      if(NoSurvey || Nsurveys==0){next}
      out=AnomalyDetectionTS(t(data.matrix(y[,3:(3+Nsurveys-1)])),B=1000,onesided); DTYPE="Surveys" #surveys only
      outmat[rIDs[c(-1,-length(rIDs))],1] = out$output[1,]
      outmat[rIDs[c(-1,-length(rIDs))],2] = out$output[2,]
      outmat[rIDs[c(-1,-length(rIDs))],3] = out$output[3,]
    }else if(i==2){
      if(Nmobfeats==0){next}
      out=AnomalyDetectionTS(t(data.matrix(y[,(3+Nsurveys):(3+Nsurveys+Nmobfeats-1)])),B=1000,onesided); DTYPE="Mobility" #mobility only
      outmat[rIDs[c(-1,-length(rIDs))],4] = out$output[1,]
      outmat[rIDs[c(-1,-length(rIDs))],5] = out$output[2,]
      outmat[rIDs[c(-1,-length(rIDs))],6] = out$output[3,]
    }else{
      if(Nsocfeats==0){next}
      out=AnomalyDetectionTS(t(data.matrix(y[,(3+Nsurveys+Nmobfeats):ncol(y)])),B=1000,onesided); DTYPE="Sociability" #sociability only
      outmat[rIDs[c(-1,-length(rIDs))],7] = out$output[1,]
      outmat[rIDs[c(-1,-length(rIDs))],8] = out$output[2,]
      outmat[rIDs[c(-1,-length(rIDs))],9] = out$output[3,]
    }
  }
  return(list(outmat,rIDs))
  #write.csv(cbind(dat,outmat),paste(output_filepath,"Results","Group","feature_matrix_anomalies.csv",sep="/"),quote=F,row.names=F)
}


daysback=2
featanom = readRDS(infilepath)[[1]]
newcolnames=c(names(featanom),c("SVY_stat","SVY_df","SVY_pval","MOB_stat","MOB_df","MOB_pval","SOC_stat","SOC_df","SOC_pval"))
featanom = cbind(featanom,matrix(NA,nrow=nrow(featanom),ncol=9))
colnames(featanom)=newcolnames

patient_names=pinfo$ID
Nsurveys=6
cat("\nPerforming anomaly detection for ",length(patient_names),"subjects:\n\n")
for(i in 1:length(patient_names)){
  cat(paste("ID: ",patient_names[i]," (",i,"/",length(patient_names),")\n",sep=""))
  vertmarks=c(NULL)
  ########## The following commented out portion is used to read in hospitalization dates from a
  ########## csv file and mark the resulting plots with a red line on this dates.
  rID=which(pinfo[,1]==patient_names[i])
  if(length(rID)>0){
    vertmarks=c(pinfo[rID[1],3])
    if(is.na(vertmarks)){
      vertmarks=c(NULL)
    }
  }
  out=anomaly_detection_plot(ID=patient_names[i],Nsurveys=Nsurveys,NoSurvey=FALSE,vertmarks=vertmarks,onesided=TRUE)
  rIDs = out[[2]]
  if(!is.null(rIDs) && length(rIDs)>3){
    outmat=out[[1]]
    featanom[rIDs[c(-1,-2,-length(rIDs))],((ncol(featanom)-8):ncol(featanom))]=outmat[rIDs[c(-1,-2,-length(rIDs))],]
  }
}
write.csv(featanom,"feature_matrix_anomalies.csv",quote=F,row.names=F)


